// Copyright 2006 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.apache.tapestry.ioc.internal;

import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.apache.tapestry.ioc.ModuleBuilderSource;
import org.apache.tapestry.ioc.ServiceDecorator;
import org.apache.tapestry.ioc.def.ContributionDef;
import org.apache.tapestry.ioc.def.DecoratorDef;
import org.apache.tapestry.ioc.def.ServiceDef;

/**
 * A module within the Tapestry IoC registry. Each Module is constructed around a corresponding
 * module builder instance; the methods and annotations of that instance define the services
 * provided by the module.
 * 
 * 
 */
public interface Module extends ModuleBuilderSource
{
    /**
     * Locates a service given a service id and the corresponding service interface type.
     * 
     * @param <T>
     * @param serviceId
     *            the fully qualified service id
     * @param serviceInterface
     *            the interface the service implements
     * @param module
     *            the module requesting the service, or null for no module; this is used for
     *            performing visibility checks
     * @return the service's proxy
     * @throws RuntimeException
     *             if the service does not exist (this is considered programmer error)
     */
    <T> T getService(String serviceId, Class<T> serviceInterface, Module module);

    /**
     * Locates the ids of all services that implement the provided service interface. The requesting
     * module is passed, so that visibility rules may be enforced.
     * 
     * @param serviceInterface
     *            the interface to search for
     * @param module
     *            the module requesting the search, or null if no module is requesting the search
     * @return a collection of strings, each a fully qualified id of a service
     */
    Collection<String> findServiceIdsForInterface(Class serviceInterface, Module module);

    /**
     * Locates all the decorators that should apply the identified service. This includes visibility
     * rules (private services may only be decorated by decorators in the same module) and other
     * filtering rules. The resulting list is ordered and from the list of
     * {@link org.apache.tapestry.ioc.def.DecoratorDef}s, a list of {@link ServiceDecorator}s is
     * returned.
     * 
     * @param serviceId
     *            identifies the service to be decorated
     * @return the ordered list of service decorators
     */
    List<ServiceDecorator> findDecoratorsForService(String serviceId);

    /**
     * Iterates over any decorator definitions defined by the module and returns those that apply to
     * the provided service definition.
     * 
     * @param serviceDef
     *            for which decorators are being assembled
     * @return set of decorators, possibly empty (but not null)
     */
    Set<DecoratorDef> findMatchingDecoratorDefs(ServiceDef serviceDef);

    /** Returns the id that uniquely identifies this module within the registry. */
    String getModuleId();

    /** Finds any contributions that are targetted at the indicated service. */
    Set<ContributionDef> getContributorDefsForService(String serviceId);

    /**
     * Locates services with the {@link org.apache.tapestry.ioc.annotations.EagerLoad} annotation
     * and forces them to instantiate fully. This is part of the Registry startup.
     */
    void eagerLoadServices();
}
